package be.cylab.java.wowa.training;

import be.cylab.java.roc.Roc;
import be.cylab.java.roc.RocCoordinates;
import info.debatty.java.aggregation.WOWA;

import java.text.DateFormat;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Date;
import java.util.List;
import java.util.Random;


/**
 * Abstract class Solution.
 */
public abstract class AbstractSolution
        implements Comparable<AbstractSolution>, Cloneable {

    protected List<Double> weights_w;
    protected List<Double> weights_p;
    protected Double fitness_score = Double.POSITIVE_INFINITY;

    /**
     * SolutionDistance constructor. Needs weight number as parameter
     *
     * @param weights_number
     */
    public AbstractSolution(final int weights_number) {
        this.weights_w = new ArrayList<>();
        this.weights_p = new ArrayList<>();
        for (int i = 0; i < weights_number; i++) {
            this.weights_w.add(Math.random());
            this.weights_p.add(Math.random());
        }
        this.normalize();
    }

    /**
     * Constructor solution distance. Made for test to initialize the Random
     * Generator
     *
     * @param weights_number int
     * @param seed           int
     */

    public AbstractSolution(final int weights_number, final int seed) {
        Random rnd = new Random(seed);
        this.weights_w = new ArrayList<>();
        this.weights_p = new ArrayList<>();
        for (int i = 0; i < weights_number; i++) {
            this.weights_w.add(rnd.nextDouble());
            this.weights_p.add(rnd.nextDouble());
        }
        this.normalize();
    }

    abstract void computeScoreTo(List<List<Double>> data,
                                 List<Double> expected);

    /**
     * @return
     */
    @Override
    public final String toString() {
        return "SolutionDistance{"
                + "weights_w=" + Arrays.toString(weights_w.toArray())
                + ", weights_p=" + Arrays.toString(weights_p.toArray())
                + ", fitness score=" + Math.abs(fitness_score)
                + '}';
    }

    /**
     * Change a random gene in a weight vector.
     * The gene is changed is the probability is higher than a random double
     *
     * @param probability
     */
    final void randomlyMutateWithProbability(final double probability) {
        double tos = Math.random();
        if (tos > probability) {
            // do nothing
            return;
        }

        //new weight value and position
        double new_weight_value = Math.random();
        int random_index = Utils.randomInteger(0, this.weights_p.size() - 1);
        //Select w or p weights
        int weight_selection = Utils.randomInteger(0, 1);
        if (weight_selection == 0) {
            this.weights_w.set(random_index, new_weight_value);
        } else {
            this.weights_p.set(random_index, new_weight_value);
        }
        this.normalize();

    }

    /**
     * @param data
     * @param expected
     * @return
     */
    final List<RocCoordinates> computeRocPoints(
            final List<List<Double>> data,
            final List<Double> expected,
            final boolean save_on_csv) {
        if (data.size() != expected.size()) {
            throw new IllegalArgumentException(
                    "Data and expected have different size");
        }
        boolean[] true_alert
                = Utils.convertExpectedToBooleanArrayTrueAlert(expected);
        double[] score = this.computeWOWAScoreWithData(data);
        Roc roc = new Roc(score, true_alert);
        List<RocCoordinates> coordinates = roc.computeRocPoints();
        if (save_on_csv) {
            DateFormat format = new SimpleDateFormat("yyyyMMdd_HH:mm:ss");
            Date date = new Date();
            be.cylab.java.roc.Utils.storeRocCoordinatesInCSVFile(coordinates,
                    "RocCoordinates_" + format.format(date) + ".csv");
        }
        return coordinates;
    }

    /**
     * Method to compute AUC with List as argument.
     * @param data
     * @param expected
     * @return
     */
    final double computeAUC(
            final List<List<Double>> data,
            final List<Double> expected) {
        if (data.size() != expected.size()) {
            throw new IllegalArgumentException(
                    "Data and expected have different size");
        }
        //Convert the array with expected values (0 or 1) to a boolean array.
        boolean[] true_alert
                = Utils.convertExpectedToBooleanArrayTrueAlert(expected);

        //For each elements in dataset, compute the wowa function.
        double[] score = this.computeWOWAScoreWithData(data);
        Roc roc = new Roc(score, true_alert);
        return roc.computeAUC();
    }

    /**
     * Method to compute AUC with filename as arugument.
     * @param filename_data
     * @param filename_expected
     * @return
     */
    public final double computeAUC(
            final String filename_data,
            final String filename_expected) {
        List<List<Double>> data
                = Utils.convertJsonToDataForTrainer(filename_data);
        List<Double> expected
                = Utils.convertJsonToExpectedForTrainer(filename_expected);
        return computeAUC(data, expected);
    }

    /**
     * @param data
     * @return
     */
    double[] computeWOWAScoreWithData(
            final List<List<Double>> data) {
        double[] score = new double[data.size()];
        double[] weights_w
                = Utils.convertListDoubleToArrayDouble(this.weights_w);
        double[] weights_p
                = Utils.convertListDoubleToArrayDouble(this.weights_p);
        WOWA wowa = new WOWA(weights_w, weights_p);
        for (int i = 0; i < data.size(); i++) {
            score[i] = wowa.aggregate(
                    Utils.convertListDoubleToArrayDouble(data.get(i)));
        }
        return score;
    }

    /**
     * @param solution
     * @return int
     */
    @Override
    public final int compareTo(final AbstractSolution solution) {
        if (Double.compare(this.getFitnessScore(),
                solution.getFitnessScore()) > 0) {
            return 1;
        } else if (Double.compare(this.getFitnessScore(),
                solution.getFitnessScore()) < 0) {
            return -1;
        } else {
            return 0;
        }
    }

    /**
     * Function to normalize SolutionDistance weights.
     * Weights must be between 0 and 1 and the sum of the weight in a vector
     * must be equal to 1
     */
    final void normalize() {
        this.weights_w = Utils.normalizeWeights(this.weights_w);
        this.weights_p = Utils.normalizeWeights(this.weights_p);
    }

    /**
     * @return
     */
    public final double getFitnessScore() {
        return fitness_score;
    }


    /**
     * @return
     */
    public final List<Double> getWeightsW() {
        return this.weights_w;
    }

    /**
     * @return
     */
    public final List<Double> getWeightsP() {
        return this.weights_p;
    }

    /**
     * @param weights_w
     */
    final void setWeightsW(final List<Double> weights_w) {
        this.weights_w = weights_w;
    }

    /**
     * @param weights_p
     */
    final void setWeightsP(final List<Double> weights_p) {
        this.weights_p = weights_p;
    }


    /**
     * Copy.
     *
     * @return
     * @throws java.lang.CloneNotSupportedException if clone is not supported
     */
    public final SolutionDistance clone() throws CloneNotSupportedException {
        return (SolutionDistance) super.clone();
    }

}
