package be.cylab.java.wowa.training;

import com.owlike.genson.GenericType;
import com.owlike.genson.Genson;
import org.apache.commons.lang3.ArrayUtils;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;

/**
 * Class Utils. Contain tools for Trainer algorithm
 */
final class Utils {

    private Utils() {

    }

    /**
     * Normalize a weights vector.
     *
     * @param weights
     * @return
     */
    static List<Double> normalizeWeights(final List<Double> weights) {
        Double sum_weight = Utils.sumListElements(weights);
        if (sum_weight == 0.0) {
            throw new IllegalArgumentException(
                    "Sum of weights must be different of 0");
        }
        List<Double> weights_normalized = new ArrayList<>();
        for (int i = 0; i < weights.size(); i++) {
            weights_normalized.add(weights.get(i) / sum_weight);
        }
        return weights_normalized;
    }

    /**
     * Find the max distance in a List of SolutionDistance.
     *
     * @param solutions
     * @return
     */
    static double findMaxDistance(
            final List<AbstractSolution> solutions) {
        double max = Double.NEGATIVE_INFINITY;
        for (AbstractSolution solution : solutions) {
            if (solution.getFitnessScore() > max) {
                max = solution.getFitnessScore();
            }
        }
        return max;
    }

    /**
     * Find the minimum distance in a List of SolutionDistance.
     *
     * @param solutions
     * @return
     */
    static double findMinDistance(
            final List<AbstractSolution> solutions) {
        double min = Double.POSITIVE_INFINITY;
        for (AbstractSolution solution : solutions) {
            if (solution.getFitnessScore() < min) {
                min = solution.getFitnessScore();
            }
        }
        return min;
    }

    /**
     * Compute the sum of all distance in a List of SolutionDistance.
     *
     * @param solutions
     * @return
     */
    static double sumTotalDistance(
            final List<AbstractSolution> solutions) {
        double sum = 0;
        for (AbstractSolution solution : solutions) {
            sum += solution.getFitnessScore();
        }
        return sum;
    }

    /**
     * Compute the sum of all array's element.
     *
     * @param array
     * @return
     */
    static Double sumListElements(final List<Double> array) {
        Double sum = 0.0;
        for (Double weight : array) {
            sum += weight;
        }
        return sum;
    }

    /**
     * Generate a random integer in a range.
     *
     * @param min
     * @param max
     * @return
     */
    static int randomInteger(final int min, final int max) {
        if (min >= max) {
            throw new IllegalArgumentException("Max must be greater then min");
        }
        Random rnd = new Random();
        return rnd.nextInt((max - min) + 1) + min;
    }

    /**
     * Import a json file.
     *
     * @param filename
     * @return
     */
    static List<List<Double>> convertJsonToDataForTrainer(
            final String filename) {
        Genson genson = new Genson();
        String data_json = Utils.readFileToString(filename);
        List<List<Double>> data = genson.deserialize(
                data_json,
                new GenericType<List<List<Double>>>() {
                });
        return data;
    }

    /**
     * Import result file.
     *
     * @param filename
     * @return
     */
    static List<Double> convertJsonToExpectedForTrainer(
            final String filename) {

        Genson genson = new Genson();
        String expected_json = Utils.readFileToString(filename);
        List<Double> expected = genson.deserialize(expected_json,
                new GenericType<List<Double>>() {
                });
        return expected;
    }

    /**
     * Method to read file and convert into a String.
     *
     * @param filename
     * @return
     */
    private static String readFileToString(final String filename) {
        String json = null;
        try {
            json = new String(Files.readAllBytes(Paths.get(filename)),
                    StandardCharsets.UTF_8);
        } catch (IOException e) {
            e.printStackTrace();
        }
        return json;
    }


    /**
     * @param expected
     * @return
     */
    static boolean[] convertExpectedToBooleanArrayTrueAlert(
            final List<Double> expected) {
        boolean[] true_alert = new boolean[expected.size()];
        for (int i = 0; i < expected.size(); i++) {
            if (expected.get(i) == 1) {
                true_alert[i] = true;
            } else if (expected.get(i) == 0) {
                true_alert[i] = false;
            } else {
                throw new IllegalStateException(
                        "Expected must be equal to 0 or 1 to compute ROC");
            }
        }
        return true_alert;
    }

    /**
     * Function to convert a List of doubles to an Array of doubles.
     *
     * @param elements
     * @return
     */
    static double[] convertListDoubleToArrayDouble(
            final List<Double> elements) {
        Double[] w = new Double[elements.size()];
        w = elements.toArray(w);
        double[] ww = ArrayUtils.toPrimitive(w);
        return ww;
    }

    /**
     * @param size
     * @return
     */
    static List<Double> initializeListWithZeroValues(
            final int size) {
        List<Double> list = new ArrayList<>();
        for (int i = 0; i < size; i++) {
            list.add(0.0);
        }
        return list;
    }

}
